unit UClientGUI;

{$mode delphi}

interface

uses
  Types, Classes, SysUtils, Graphics, UGameClient, UPlayer, UMap, UGame, UGeometry,
  Math;

type
  { TView }

  TView = class
  private
    FDestRect: TRect;
    FZoom: Double;
    procedure SetDestRect(AValue: TRect);
    procedure SetZoom(AValue: Double);
  public
    Game: TObject; //TGame;
    SourceRect: TRect;
    FocusedCell: TPlayerCell;
    SelectedCell: TPlayerCell;
    procedure Clear;
    function IsCellVisible(Cell: TCell): Boolean;
    constructor Create;
    destructor Destroy; override;
    procedure SelectCell(Pos: TPoint; Player: TPlayer; ShiftState: TShiftState);
    procedure CenterMap;
    procedure CenterPlayerCity(Player: TPlayer);
    function CanvasToCellPos(Pos: TPoint): TPoint;
    function CellToCanvasPos(Pos: TPoint): TPoint;
    function CanvasToCellRect(Pos: TRect): TRect;
    function CellToCanvasRect(Pos: TRect): TRect;
    procedure Assign(Source: TView);
    property DestRect: TRect read FDestRect write SetDestRect;
    property Zoom: Double read FZoom write SetZoom;
  end;

  { TClientGUI }

  TClientGUI = class(TClient)
  protected
    procedure SetGame(AValue: TGame); override;
  public
    View: TView;
    procedure PaintCell(Canvas: TCanvas; Pos: TPoint; Text: string; View: TView;
      Cell: TPlayerCell);
    procedure PaintMapCell(Canvas: TCanvas; Pos: TPoint; Text: string; View: TView;
      Cell: TCell);
    procedure Paint(Canvas: TCanvas; View: TView);
    constructor Create; override;
    destructor Destroy; override;
  end;

implementation

resourcestring
  SZeroZoomNotAlowed = 'Zero zoom not allowed';
  SWrongArrowAngle = 'Wrong arrow angle %s';

{ TClientGUI }

procedure TClientGUI.Paint(Canvas: TCanvas; View: TView);
var
  I: Integer;
  Cell: TPlayerCell;
  MapCell: TCell;
  PosFrom, PosTo: TPoint;
  Angle: Double;
  ArrowCenter: TPoint;
  Move: TUnitMove;
  CellText: string;
  CellLink: TCellLink;
  NeighCell: TCell;
begin
  with Canvas, View do
  try
    Lock;
    // Draw cell links
    Pen.Color := clBlack;
    Pen.Style := psSolid;
    Pen.Width := 3;
    for CellLink in TGame(Game).Map.CellLinks do
    with CellLink do begin
      if Length(Points) >= 2 then begin
        MoveTo(PointToStdPoint(View.CellToCanvasPos(Points[0])));
        for I := 1 to Length(Points) - 1 do
          LineTo(PointToStdPoint(View.CellToCanvasPos(Points[I])));
      end;
    end;

    // Draw cells
    if Assigned(ControlPlayer) then begin
      for Cell in ControlPlayer.PlayerMap.Cells do begin
        if View.IsCellVisible(Cell.MapCell) and (Cell.MapCell.Terrain <> ttVoid) then begin
          if Cell.MapCell.Player = ControlPlayer then
            CellText := IntToStr(Cell.GetAvialPower)
            else CellText := IntToStr(Cell.MapCell.Power);
          if Assigned(SelectedCell) and (SelectedCell = Cell) then
              Brush.Color := clGreen
            else if Assigned(SelectedCell) and TGame(Game).Map.IsCellsNeighbor(SelectedCell.MapCell, Cell.MapCell) then
              Brush.Color := clPurple
            else if TGame(Game).FogOfWar then begin
              if Cell.InVisibleRange then begin
                Brush.Color := Cell.MapCell.GetColor;
              end else begin
                if Cell.Explored then begin
                  Brush.Color := $404040;
                  CellText := '';
                end else begin
                  Brush.Color := clBlack;
                  CellText := '';
                end;
              end;
            end else Brush.Color := Cell.MapCell.GetColor;
          PaintCell(Canvas, Cell.MapCell.PosPx, CellText, View, Cell);
        end else
        if TGame(Game).FogOfWar and (Cell.MapCell.Terrain = ttVoid) then begin
          Brush.Color := clBlack;
          PaintCell(Canvas, Cell.MapCell.PosPx, '', View, Cell);
        end;
      end;
    end else begin
      // Draw cells
      for MapCell in TGame(Game).Map.Cells do begin
        if (MapCell.Terrain <> ttVoid) and View.IsCellVisible(MapCell) then begin
          Brush.Color := MapCell.GetColor;
          PaintMapCell(Canvas, MapCell.PosPx, IntToStr(MapCell.Power), View, MapCell);
        end;
      end;
    end;

    // Draw links to neighbors
    if TGame(Game).DevelMode then
    for Cell in ControlPlayer.PlayerMap.Cells do begin
      for NeighCell in Cell.MapCell.Neighbors do begin
        Pen.Color := clYellow;
        MoveTo(PointToStdPoint(View.CellToCanvasPos(Cell.MapCell.PosPx)));
        LineTo(PointToStdPoint(View.CellToCanvasPos(NeighCell.PosPx)));
      end;

      Font.Color := clRed;
      Brush.Style := bsClear;
      TextOut(View.CellToCanvasPos(Cell.MapCell.PosPx).X,
        View.CellToCanvasPos(Cell.MapCell.PosPx).Y, IntToStr(Cell.MapCell.Id));
    end;

    // Draw arrows
    Pen.Color := clCream;
    if Assigned(ControlPlayer) then
    for Move in ControlPlayer.Moves do begin
      PosFrom := TGame(Game).Map.CellToPos(Move.CellFrom.MapCell);
      PosTo := TGame(Game).Map.CellToPos(Move.CellTo.MapCell);
      // In Fog of war mode show only
      if TGame(Game).FogOfWar and not ControlPlayer.PlayerMap.Cells.SearchCell(Move.CellFrom.MapCell).InVisibleRange and
        not ControlPlayer.PlayerMap.Cells.SearchCell(Move.CellTo.MapCell).InVisibleRange then
        Continue;
      if Move.CountRepeat > 0 then Pen.Width := 2
        else Pen.Width := 1;
      Angle := ArcTan((PosTo.Y - PosFrom.Y) / (PosTo.X - PosFrom.X));
      if (Angle > +Pi) or (Angle < -Pi) then
        raise Exception.Create(Format(SWrongArrowAngle, [FloatToStr(Angle)]));

      if Sign(PosTo.X - PosFrom.X) = -1 then Angle := Angle + Pi;
      ArrowCenter := View.CellToCanvasPos(TPoint.Create(Trunc(PosFrom.X + (PosTo.X - PosFrom.X) / 2),
        Trunc(PosFrom.Y + (PosTo.Y - PosFrom.Y) / 2)));
      TGame(Game).Map.DrawArrow(Canvas, ArrowCenter,
        Angle, IntToStr(Move.CountOnce), View.Zoom);
    end;
  finally
    Unlock;
  end;
end;

constructor TClientGUI.Create;
begin
  inherited;
  View := TView.Create;
end;

destructor TClientGUI.Destroy;
begin
  FreeAndNil(View);
  inherited Destroy;
end;

procedure TClientGUI.SetGame(AValue: TGame);
begin
  inherited;
  View.Game := AValue;
end;

procedure TClientGUI.PaintCell(Canvas: TCanvas; Pos: TPoint; Text: string;
  View: TView; Cell: TPlayerCell);
var
  I: Integer;
  TextPos: TPoint;
  Points: array of Classes.TPoint;
  TextSize: TSize;
begin
  if Cell.MapCell.Extra = etObjectiveTarget then begin
    Text := Text + '!';
  end;
  with Canvas do begin
    if Assigned(View.FocusedCell) and (View.FocusedCell = Cell) then begin
      Pen.Color := clYellow;
      Pen.Style := psSolid;
      Pen.Width := 1;
    end else
    if Cell.MapCell.Terrain = ttCity then begin
      // Cannot set clear border as it will display shifted on gtk2
      //Pen.Style := psClear;
      Pen.Color := clBlack;
      Pen.Style := psSolid;
      Pen.Width := 3;
    end else begin
      // Cannot set clear border as it will display shifted on gtk2
      //Pen.Style := psClear;
      Pen.Color := Brush.Color;
      Pen.Style := psSolid;
      Pen.Width := 0;
    end;
    // Transform view
    SetLength(Points, Length(Cell.MapCell.Polygon.Points));
    for I := 0 to Length(Points) - 1 do
      Points[I] := PointToStdPoint(View.CellToCanvasPos(Cell.MapCell.Polygon.Points[I]));
    Brush.Style := bsSolid;
    //Polygon(Points, False, 0, Length(Points));
    TCanvasEx.PolygonEx(Canvas, Points, False);
    //MoveTo(Points[0].X, Points[0].Y);
    //LineTo(Points[1].X, Points[1].Y);

    // Show cell text
    if (Cell.GetAvialPower <> 0) or (Cell.MapCell.Extra = etObjectiveTarget) then begin
      Pen.Style := psSolid;
      Font.Color := clWhite;
      Brush.Style := bsClear;
      Font.Size := Trunc(42 * View.Zoom);
      TextPos := View.CellToCanvasPos(Pos);
      TextSize := TextExtent(Text);
      TCanvasEx.TextOutEx(Canvas, Round(TextPos.X) - TextSize.cx div 2,
        Round(TextPos.Y) - TextSize.cy div 2, Text, False);
    end;
  end;
end;

procedure TClientGUI.PaintMapCell(Canvas: TCanvas; Pos: TPoint; Text: string;
  View: TView; Cell: TCell);
var
  I: Integer;
  TextPos: TPoint;
  Points: array of Classes.TPoint;
  TextSize: TSize;
begin
  if Cell.Extra = etObjectiveTarget then begin
    Text := Text + '!';
  end;
  with Canvas do begin
    if Assigned(View.FocusedCell) and (View.FocusedCell.MapCell = Cell) then begin
      Pen.Color := clYellow;
      Pen.Style := psSolid;
      Pen.Width := 1;
    end else
    if Cell.Terrain = ttCity then begin
      // Cannot set clear border as it will display shifted on gtk2
      //Pen.Style := psClear;
      Pen.Color := clBlack;
      Pen.Style := psSolid;
      Pen.Width := 3;
    end else begin
      // Cannot set clear border as it will display shifted on gtk2
      //Pen.Style := psClear;
      Pen.Color := Brush.Color;
      Pen.Style := psSolid;
      Pen.Width := 0;
    end;
    // Transform view
    SetLength(Points, Length(Cell.Polygon.Points));
    for I := 0 to Length(Points) - 1 do
      Points[I] := PointToStdPoint(View.CellToCanvasPos(Cell.Polygon.Points[I]));
    Brush.Style := bsSolid;
    //Polygon(Points, False, 0, Length(Points));
    TCanvasEx.PolygonEx(Canvas, Points, False);
    //MoveTo(Points[0].X, Points[0].Y);
    //LineTo(Points[1].X, Points[1].Y);

    // Show cell text
    if (Cell.Power <> 0) or (Cell.Extra = etObjectiveTarget) then begin
      Pen.Style := psSolid;
      Font.Color := clWhite;
      Brush.Style := bsClear;
      Font.Size := Trunc(42 * View.Zoom);
      TextPos := View.CellToCanvasPos(Pos);
      TextSize := TextExtent(Text);
      TCanvasEx.TextOutEx(Canvas, Round(TextPos.X) - TextSize.cx div 2,
        Round(TextPos.Y) - TextSize.cy div 2, Text, False);
    end;
  end;
end;

{ TView }

procedure TView.SetZoom(AValue: Double);
begin
  if FZoom = AValue then Exit;
  if AValue = 0 then
    raise Exception.Create(SZeroZoomNotAlowed);
  FZoom := AValue;
  SourceRect := TRect.CreateBounds(TPoint.Create(Trunc(SourceRect.P1.X + SourceRect.Size.X div 2 - DestRect.Size.X / Zoom / 2),
    Trunc(SourceRect.P1.Y +  SourceRect.Size.Y div 2 - DestRect.Size.Y / Zoom / 2)),
    TPoint.Create(Trunc(DestRect.Size.X / Zoom),
    Trunc(DestRect.Size.Y / Zoom)));
end;

procedure TView.Clear;
begin
  FocusedCell := nil;
  SelectedCell := nil;
end;

procedure TView.SetDestRect(AValue: TRect);
var
  Diff: TPoint;
begin
  if FDestRect = AValue then Exit;
  Diff := TPoint.Create(Trunc(DestRect.Size.X / Zoom - AValue.Size.X / Zoom) div 2,
    Trunc(DestRect.Size.Y / Zoom - AValue.Size.Y / Zoom) div 2);
  FDestRect := AValue;
  SourceRect := TRect.CreateBounds(TPoint.Create(SourceRect.P1.X + Diff.X, SourceRect.P1.Y + Diff.Y),
    TPoint.Create(Trunc(DestRect.Size.X / Zoom),
    Trunc(DestRect.Size.Y / Zoom)));
end;

constructor TView.Create;
begin
  Zoom := 1.5;
  Clear;
end;

destructor TView.Destroy;
begin
  inherited Destroy;
end;

function TView.CanvasToCellPos(Pos: TPoint): TPoint;
begin
  Result := TPoint.Create(Trunc((Pos.X - DestRect.P1.X) / Zoom + SourceRect.P1.X),
    Trunc((Pos.Y - DestRect.P1.Y) / Zoom + SourceRect.P1.Y));
end;

function TView.CellToCanvasPos(Pos: TPoint): TPoint;
begin
  Result := TPoint.Create(Trunc((Pos.X - SourceRect.P1.X) * Zoom) + DestRect.P1.X,
    Trunc((Pos.Y - SourceRect.P1.Y) * Zoom) + DestRect.P1.Y);
end;

function TView.CanvasToCellRect(Pos: TRect): TRect;
begin
  Result.P1 := CanvasToCellPos(Pos.P1);
  Result.P2 := CanvasToCellPos(Pos.P2);
end;

function TView.CellToCanvasRect(Pos: TRect): TRect;
begin
  Result.P1 := CellToCanvasPos(Pos.P1);
  Result.P2 := CellToCanvasPos(Pos.P2);
end;

procedure TView.Assign(Source: TView);
begin
  SourceRect := Source.SourceRect;
  FDestRect := Source.DestRect;
  FZoom := Source.Zoom;
  SelectedCell := Source.SelectedCell;
  FocusedCell := Source.FocusedCell;
end;

procedure TView.SelectCell(Pos: TPoint; Player: TPlayer; ShiftState: TShiftState);
var
  NewSelectedCell: TPlayerCell;
  UnitMove: TUnitMove;
  I: Integer;
  CellPos: TPoint;
  R: TRect;
begin
  if TGame(Game).Map.Cyclic then begin
    R := CellToCanvasRect(TGame(Game).Map.PixelRect);
    CellPos := TPoint.Create(
      ModNeg(Pos.X - R.P1.X, R.Size.X) + R.P1.X,
      ModNeg(Pos.Y - R.P1.Y, R.Size.Y) + R.P1.Y
    );
    NewSelectedCell := Player.PlayerMap.PosToCell(
      CanvasToCellPos(CellPos));
  end else begin
    NewSelectedCell := Player.PlayerMap.PosToCell(
      CanvasToCellPos(Pos));
  end;
  if Assigned(NewSelectedCell) then begin
    if Assigned(SelectedCell) and TGame(Game).Map.IsCellsNeighbor(NewSelectedCell.MapCell, SelectedCell.MapCell) then begin
      if ssShift in ShiftState then begin
        // Make maximum unit move without confirmation dialog
        for I := SelectedCell.MovesFrom.Count - 1 downto 0 do begin
          Player.Moves.Remove(SelectedCell.MovesFrom[I]);
        end;
        TGame(Game).CurrentPlayer.SetMove(SelectedCell, NewSelectedCell, SelectedCell.MapCell.Power, False);
        SelectedCell := nil;
      end else
      if ssCtrl in ShiftState then begin
        // If CTRL key pressed then storno all moved from selected cell and
        // move all power to new selected cell
        for I := SelectedCell.MovesFrom.Count - 1 downto 0 do
          Player.Moves.Remove(SelectedCell.MovesFrom[I]);
        UnitMove := TGame(Game).CurrentPlayer.SetMove(SelectedCell, NewSelectedCell, SelectedCell.MapCell.Power, False);
        if Assigned(UnitMove) then
          UnitMove.CountRepeat := TGame(Player.Game).Map.MaxPower;
        if NewSelectedCell.MapCell.Player = Player then SelectedCell := NewSelectedCell
          else SelectedCell := nil;
      end else begin
        TGame(Game).CurrentPlayer.SetMove(SelectedCell, NewSelectedCell, SelectedCell.MapCell.Power);
        SelectedCell := nil;
      end;
    end else
    if (NewSelectedCell <> SelectedCell) and (NewSelectedCell.MapCell.Player = Player) then
      SelectedCell := NewSelectedCell
    else
    if (NewSelectedCell = SelectedCell) and (NewSelectedCell.MapCell.Player = Player) then
      SelectedCell := nil;
  end;
end;

procedure TView.CenterMap;
var
  MapRect: TRect;
begin
  MapRect := TGame(Game).Map.PixelRect;
  SourceRect := TRect.CreateBounds(TPoint.Create(MapRect.P1.X + MapRect.Size.X div 2 - SourceRect.Size.X div 2,
    MapRect.P1.Y + MapRect.Size.Y div 2 - SourceRect.Size.Y div 2),
    TPoint.Create(SourceRect.Size.X,
    SourceRect.Size.Y));
end;

procedure TView.CenterPlayerCity(Player: TPlayer);
begin
  SourceRect := TRect.CreateBounds(TPoint.Create(Player.StartCell.PosPx.X - SourceRect.Size.X div 2,
    Player.StartCell.PosPx.Y - SourceRect.Size.Y div 2),
    TPoint.Create(SourceRect.Size.X,
    SourceRect.Size.Y));
end;

function TView.IsCellVisible(Cell: TCell): Boolean;
var
  RectPolygon, RectView: TRect;
begin
  RectPolygon := Cell.Polygon.GetRect;
  RectView := SourceRect;
  Result := (
    (RectPolygon.P1.X < RectView.P2.X) and
    (RectPolygon.P2.X > RectView.P1.X) and
    (RectPolygon.P1.Y < RectView.P2.Y) and
    (RectPolygon.P2.Y > RectView.P1.Y)
  );
  Result := True;
end;

end.

